/*****************************************************************************\
 *  bridge_linker.c
 *
 *****************************************************************************
 *  Copyright (C) 2004 The Regents of the University of California.
 *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
 *  Written by Dan Phung <phung4@llnl.gov>, Danny Auble <da@llnl.gov>
 *
 *  This file is part of SLURM, a resource management program.
 *  For details, see <https://computing.llnl.gov/linux/slurm/>.
 *  Please also read the included file: DISCLAIMER.
 *
 *  SLURM is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  In addition, as a special exception, the copyright holders give permission
 *  to link the code of portions of this program with the OpenSSL library under
 *  certain conditions as described in each individual source file, and
 *  distribute linked combinations including the two. You must obey the GNU
 *  General Public License in all respects for all of the code used other than
 *  OpenSSL. If you modify file(s) with this exception, you may extend this
 *  exception to your version of the file(s), but you are not obligated to do
 *  so. If you do not wish to do so, delete this exception statement from your
 *  version.  If you delete this exception statement from all source files in
 *  the program, then also delete it here.
 *
 *  SLURM is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with SLURM; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA.
\*****************************************************************************/


#include "bridge_linker.h"

#if defined HAVE_BG_FILES && defined HAVE_BG_Q
#ifdef __cplusplus
extern "C" {
#endif

using namespace bgsched;

pthread_mutex_t api_file_mutex = PTHREAD_MUTEX_INITIALIZER;
bridge_api_t bridge_api;
bool initialized = false;
bool have_db2 = true;
void *handle = NULL;

extern int bridge_init()
{
	if (initialized)
		return 1;

	bgsched::init(properties);


	initialized = true;

	return 1;

}

extern int bridge_fini()
{
	initialized = false;

	return SLURM_SUCCESS;
}

extern status_t bridge_get_bg(my_bluegene_t **bg)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.get_bg))(bg);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;
}

extern status_t bridge_add_block(rm_partition_t *partition)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.add_partition))(partition);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_get_block(pm_partition_id_t pid,
				 rm_partition_t **partition)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.get_partition))(pid, partition);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_get_block_info(pm_partition_id_t pid,
				      rm_partition_t **partition)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	/* this is here to make sure we don't lock up things with
	   polling and the long running get_BG call */
	rc = pthread_mutex_trylock(&api_file_mutex);
	if (rc == EBUSY)
		return rc;
	else if (rc) {
		errno = rc;
		error("%s:%d %s: pthread_mutex_trylock(): %m",
		      __FILE__, __LINE__, __CURRENT_FUNC__);
	}

	//slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.get_partition_info))(pid, partition);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_modify_block(pm_partition_id_t pid,
				    enum rm_modify_op op, const void *data)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.modify_partition))(pid, op, data);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_set_block_owner(pm_partition_id_t pid, const char *name)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.set_part_owner))(pid, name);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_add_block_user(pm_partition_id_t pid, const char *name)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.add_part_user))(pid, name);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_remove_block_user(pm_partition_id_t pid,
					 const char *name)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.remove_part_user))(pid, name);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_remove_block(pm_partition_id_t pid)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.remove_partition))(pid);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_get_blocks(rm_partition_state_flag_t flag,
				  rm_partition_list_t **part_list)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.get_partitions))(flag, part_list);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_get_blocks_info(rm_partition_state_flag_t flag,
				       rm_partition_list_t **part_list)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.get_partitions_info))(flag, part_list);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_get_job(db_job_id_t dbJobId, rm_job_t **job)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.get_job))(dbJobId, job);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_get_jobs(rm_job_state_flag_t flag, rm_job_list_t **jobs)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.get_jobs))(flag, jobs);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_remove_job(db_job_id_t jid)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.remove_job))(jid);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_get_nodecards(rm_bp_id_t bpid,
				     rm_nodecard_list_t **nc_list)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.get_nodecards))(bpid, nc_list);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_new_nodecard(rm_nodecard_t **nodecard)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.new_nodecard))(nodecard);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_free_nodecard(rm_nodecard_t *nodecard)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.free_nodecard))(nodecard);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

#ifndef HAVE_BGL
extern status_t bridge_new_ionode(rm_ionode_t **ionode)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.new_ionode))(ionode);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_free_ionode(rm_ionode_t *ionode)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.free_ionode))(ionode);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}
#endif

extern status_t bridge_new_block(rm_partition_t **partition)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.new_partition))(partition);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_free_block(rm_partition_t *partition)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.free_partition))(partition);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_free_job(rm_job_t *job)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.free_job))(job);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_free_bg(my_bluegene_t *bg)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.free_bg))(bg);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_free_block_list(rm_partition_list_t *part_list)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.free_partition_list))(part_list);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_free_job_list(rm_job_list_t *job_list)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.free_job_list))(job_list);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_free_nodecard_list(rm_nodecard_list_t *nc_list)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.free_nodecard_list))(nc_list);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_get_data(rm_element_t* element,
				enum rm_specification field, void *data)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.get_data))(element, field, data);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_set_data(rm_element_t* element,
				enum rm_specification field, void *data)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.set_data))(element, field, data);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

/* all the jm functions */
extern status_t bridge_signal_job(db_job_id_t jid, rm_signal_t sig)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.signal_job))(jid, sig);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern status_t bridge_cancel_job(db_job_id_t jid)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.cancel_job))(jid);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

/* all the pm functions */
extern status_t bridge_create_block(pm_partition_id_t pid)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.create_partition))(pid);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

#ifndef HAVE_BGL
extern status_t bridge_reboot_block(pm_partition_id_t pid)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.reboot_partition))(pid);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}
#endif

extern status_t bridge_destroy_block(pm_partition_id_t pid)
{
	int rc = CONNECTION_ERROR;
	if (!bridge_init())
		return rc;

	slurm_mutex_lock(&api_file_mutex);
	rc = (*(bridge_api.destroy_partition))(pid);
	slurm_mutex_unlock(&api_file_mutex);
	return rc;

}

extern int bridge_set_log_params(char *api_file_name, unsigned int level)
{
	static FILE *fp = NULL;
        FILE *fp2 = NULL;
	int rc = SLURM_SUCCESS;

	if (!bridge_init())
		return SLURM_ERROR;

	slurm_mutex_lock(&api_file_mutex);
	if (fp)
		fp2 = fp;

	fp = fopen(api_file_name, "a");

	if (fp == NULL) {
		error("can't open file for bridgeapi.log at %s: %m",
		      api_file_name);
		rc = SLURM_ERROR;
		goto end_it;
	}


	(*(bridge_api.set_log_params))(fp, level);
	/* In the libraries linked to from the bridge there are stderr
	   messages send which we would miss unless we dup this to the
	   log */
	//(void)dup2(fileno(fp), STDERR_FILENO);

	if (fp2)
		fclose(fp2);
end_it:
	slurm_mutex_unlock(&api_file_mutex);
	return rc;
}

#ifdef __cplusplus
}
#endif

#endif /* HAVE_BG_FILES */


